module DKIDE.Project (
	Project(..),
	defaultProject,
	modifyCurrentOpenFile,
	getFullProjectTree,
	addProjectFile,
	removeProjectFile,
	modifyIdeArgs,
	addLibrary,
	removeLibrary,
	getFilePath,
	CurrentProject,
	initCurrentProject,
	modifyCurrentProject,
	replaceCurrentProject,
	loadCurrentProject,
	debugProject,
	modifyGppArgs,
	modifyLdArgs
) where

-- DKIDE
import DKIDE.Parse
-- System imports
import System.IO as IO
-- Data imports
import Data.List as L
import Data.Tree
import Data.Map.Lazy as ML
import Text.Read
import Data.IORef

data Project = Project {
	projectName :: String,
	projectPath :: String,					-- Prefix pro project files
	projectFiles :: [String],
	libraryFiles :: [Tree String],	-- Kořenový uzel je prefix pro všechny cesty jeho dceřiných uzlů
	currentOpenFile :: [Int],				-- List s jedním prvkem je index do project files, s 2 hodnotamy je index do stromu library files
	gppArgs :: [String],
	ldArgs :: [String],
	qemuArgs :: [String],
	ideArgs :: ML.Map String String
} deriving(Eq,Show,Read)

debugProject :: Project -> String
debugProject proj =
	"Project dumb ---------" ++ "\n" ++
	"Project name: " ++ projectName proj ++ "\n" ++
	"Project path: " ++ projectPath proj ++ "\n" ++
	"Project files: " ++ show (projectFiles proj) ++ "\n" ++
	"Project library files: \n" ++ L.init (L.foldl (\old t-> old ++ drawTree t ++ "\n") [] (libraryFiles proj)) ++ 
	"Project open file: " ++ show (currentOpenFile proj) ++ "\n" ++
	"G++ args: " ++ show (gppArgs proj) ++ "\n" ++
	"LD args: " ++ show (ldArgs proj) ++ "\n" ++
	"Qemu args: " ++ show (qemuArgs proj) ++ "\n" ++
	"IDE args: " ++ show (ML.toList (ideArgs proj)) ++ "\n ------------- \n"

exampleProject :: Project 
exampleProject =
	Project "Delos" "/secret/hidden/Delos/Kernel" ["delos.ldscript","entry.s","kmain.cpp","kmain.h"]
	[Node "/secret/hidden/Delos/FrameWorks/Nectar" [Node "dynarray.cpp" [],Node "hashTable.cpp" []],
	Node "/secret/hidden/Delos/FrameWorks/Bloom" [Node "gdt.cpp" [],Node "idt.cpp" []]] [1,0] ["-O 4","-nostdlib"] ["-T delos.ldscript"] ["-m 64","-a a.out"]
	(ML.fromList [("style","white"),("showbar","up"),("linecoloring","yes")])

defaultProject :: Project
defaultProject =
	Project [] [] [] [] [] [] [] [] ML.empty

modifyCurrentOpenFile :: ([Int] -> [Int]) -> Project -> Project
modifyCurrentOpenFile fun (Project pn pp pf lf opfile g l q i) =
	Project pn pp pf lf (fun opfile) g l q i
	
modifyIdeArgs :: (ML.Map String String -> ML.Map String String) -> Project -> Project
modifyIdeArgs fun (Project pn pp pf lf opfile g l q i) =
	Project pn pp pf lf opfile g l q (fun i)

modifyGppArgs :: ([String] -> [String]) -> Project -> Project
modifyGppArgs fun (Project pn pp pf lf opfile g l q i) =
	Project pn pp pf lf opfile (fun g) l q i
	
modifyLdArgs :: ([String] -> [String]) -> Project -> Project
modifyLdArgs fun (Project pn pp pf lf opfile g l q i) =
	Project pn pp pf lf opfile g (fun l) q i

getFullProjectTree :: Project -> Tree String
getFullProjectTree proj = 
	Node (projectName proj) (L.map (\x -> Node x []) (projectFiles proj) ++ L.map (\(Node d c) -> Node (nameFromPath d) c) (libraryFiles proj))

addProjectFile :: String -> Project -> Project
addProjectFile name (Project pn pp pf lf opfile g l q i) =
	Project pn pp (pf ++ [name]) lf opfile g l q i

removeProjectFile :: String -> Project -> Project
removeProjectFile name (Project pn pp pf lf opfile g l q i) =
	Project pn pp (L.filter (\f -> f /= name) pf) lf opfile g l q i
	
addLibrary :: Tree String -> Project -> Project
addLibrary lib (Project pn pp pf lf opfile g l q i) =
	Project pn pp pf (lf ++ [lib]) opfile g l q i

-- Jméno bez cesty
removeLibrary :: String -> Project -> Project
removeLibrary name (Project pn pp pf lf opfile g l q i) =
	Project pn pp pf (L.filter (\(Node n _) -> nameFromPath n /= name) lf) opfile g l q i

getFilePath :: [Int] -> Project -> String
getFilePath [] _ = []
getFilePath (_:[]) _ = []
getFilePath (_:x:[]) (Project _ pp pf lf _ _ _ _ _) = 
	if L.length pf <= x
	then
		case lf !! (x - L.length pf) of
			Node v _ -> v
	else
		pp ++ "/" ++ (pf !! x)
getFilePath (_:x:y:[]) (Project _ _ pf lf _ _ _ _ _) =
		base ++ "/" ++ f
	where
		(Node base childs) = lf !! x 
		(Node f _) = childs !! y
getFilePath _ _ = []	

type CurrentProject = IORef Project

initCurrentProject :: String -> IO CurrentProject
initCurrentProject path =
	newIORef (Project (removeDotType.nameFromPath $ path) (purePath path) [] [] [] [] [] [] ML.empty) 
	
modifyCurrentProject :: (Project -> Project) -> CurrentProject -> IO ()
modifyCurrentProject modifyFun proj =
	modifyIORef proj modifyFun >>
	readIORef proj >>= (\p ->
		let ppath = projectPath p ++ "/" ++ projectName p ++ ".dkide"
		in IO.writeFile ppath (show p)) 

replaceCurrentProject :: String -> CurrentProject -> IO String
replaceCurrentProject path proj =
	IO.readFile path >>= (\f ->
		case readEither f :: Either String Project of
			Right (Project _ _ pfiles lfiles opfile g l q i) -> 
				modifyIORef proj (\_ -> Project (removeDotType.nameFromPath $ path) (purePath path) pfiles lfiles opfile g l q i) >>
				return ""
			Left err ->  
				return err)	

loadCurrentProject :: CurrentProject -> IO Project
loadCurrentProject curr =
	readIORef curr
