module Rose.Parse (
	parseLiteral,
	parseVar,
	parseCall,
	parseLocalBinds,
	parseCase,
	parseRExpr,
	comment,
	parseFunction,
	parseRose,
	parseRoseExpr,
	parseRString,
	parseRoseData
) where

import Data.List as L
import Data.Map.Lazy as ML
import Text.ParserCombinators.Parsec
import Text.Parsec.Pos
import Text.Parsec.Char

import Rose.Types
import Rose.Misc

import Debug.Trace

infixr 2 <||>
(<||>) = (<|>) 

infixr 3 <??>
(<??>) = (<?>)

comment :: GenParser Char () Char 
comment = 
	string "{-" >> 
	manyTill (try comment <|> anyChar) (try (string "-}")) >> 
	return ' ' 

blank = try comment <|> try space <|> try tab <|> try endOfLine 

parseRFloat =
	skipMany blank >>
	try (char '-' >> many1 digit >>= (\x0 ->
		char '.' >>
		many1 digit >>= (\x1 ->
			return (Literal (RFloat (read ("-" ++ x0 ++ '.' : x1))))))) <||>
	(many1 digit >>= (\x0 ->
		char '.' >>
		many1 digit >>= (\x1 ->
			return (Literal (RFloat (read (x0 ++ '.' : x1)))))))

parseRInt =
	skipMany blank >>
	try (char '-' >> many1 digit >>= (\x ->
		return (Literal (RInt (read ("-" ++ x)))))) <||>
	(many1 digit >>= (\x -> 
		return (Literal (RInt (read x))))) 

parseRBool =
	skipMany blank >>
	string "True" <||> string "False" >>= (\x -> 
		return (Literal (RBool (x == "True"))))

parseRChar = 
	skipMany blank >>
	char '\'' >>
	manyTill anyChar (char '\'') >>= (\x ->
		case x of
			"\\n" -> return (Literal (RChar '\n'))
			"\\t" -> return (Literal (RChar '\t'))
			_ -> return (Literal (RChar (x !! 0))))

parseRString =
		skipMany blank >>
		char '"' >>
		manyTill anyChar (char '"') >>= (\cs ->
			return (Literal (RList (L.map (\c -> RChar c) (map2 (\x y -> 
				if x == '\\' && y == 'n'
				then ['\n']
				else if x == '\\' && y == 't'
				then ['\t']
				else [x,y]) cs)))))

parseRUnit =
	skipMany blank >>
	char '_' >> return (Literal RUnit)
	
parseRTuple =
	skipMany blank >>
	char '(' >> sepBy parseRExpr (char ',') >>= (\x ->
		char ')' >> getPosition >>= (\p -> 
			return (Call (Var ":t-construct" (REPos (sourceName p) (sourceLine p) (sourceColumn p))) x 
			(REPos (sourceName p) (sourceLine p) (sourceColumn p)))))
	
parseRList =
	skipMany blank >>
	char '[' >> sepBy parseRExpr (char ',') >>= (\x ->
		char ']' >> getPosition >>= (\p -> 
			return (Call (Var ":l-construct" (REPos (sourceName p) (sourceLine p) (sourceColumn p))) x 
			(REPos (sourceName p) (sourceLine p) (sourceColumn p)))))
			
parseRMap =
	skipMany blank >>
	try (string "[:]" >> return (Literal (RMap [])))  <||>
	(char '[' >> sepBy (parseRExpr >>= (\f -> skipMany blank >> char ':' >> parseRExpr >>= (\s -> return [f,s]))) (char ',') >>= (\x ->
		char ']' >> getPosition >>= (\p -> 
			return (Call (Var ":m-construct" (REPos (sourceName p) (sourceLine p) (sourceColumn p))) (L.concat x) 
			(REPos (sourceName p) (sourceLine p) (sourceColumn p))))))

parseRLambda =
	skipMany blank >>
	char '\\' >> skipMany blank >> char '(' >> sepBy (skipMany blank >> many1 (noneOf "(){}[]=;:, ")) (skipMany blank >> char ',') >>= (\x ->
		skipMany blank >> char ')' >> skipMany blank >> char '=' >> parseRExpr >>= (\x1 ->
		skipMany blank >> char ';' >> return (Literal (RLambda x x1 ML.empty))))
		
parseLiteral :: GenParser Char () RExpr
parseLiteral = 
	try parseRFloat 		<|>
	try parseRInt  			<|>
	try parseRBool  		<|>
	try parseRChar  		<|>
	try parseRString 		<|>
	try parseRUnit  		<|>
	try parseRTuple 		<|>
	try parseRList 			<|>
	try parseRMap 			<|>
	parseRLambda

parseRoseData str =
		case parse parseLiteral "" str of
			Right x -> traverse x
			Left x -> RUnit
	where
		traverse (Call (Var ":l-construct" _) x _) = 
			RList (L.map traverse x)
		traverse (Call (Var ":t-construct" _) x _) = 
			RTuple (L.map traverse x)
		traverse (Call (Var ":m-construct" _) x _) = 
				RMap (L.zipWith (\f s -> (f,s)) fhalf shalf)
			where
				fhalf = L.map traverse (taker True x)
				shalf = L.map traverse (taker False x)
				taker _ [] = []
				taker True (x:xs) = x : taker False xs
				taker False (_:xs) = taker True xs
		traverse (Literal x) = x

parseVar :: GenParser Char () RExpr
parseVar =
	skipMany blank >>
	getPosition >>= (\p ->
		many1 (try (noneOf "(){}[]=;:,\t\n ")) >>= (\x -> return (Var x (REPos (sourceName p) (sourceLine p) (sourceColumn p)))))

parseCall :: GenParser Char () RExpr
parseCall =
		skipMany blank >>
		getPosition >>= (\p ->
			(try parseRLambda <|> parseVar) >>= (\lambdaExpr ->
				skipMany blank >>
				char '(' >>
				sepBy parseRExpr (skipMany blank >> char ',') >>= (\args ->
					skipMany blank >>
					char ')' >>
					chainCalls (Call lambdaExpr args (REPos (sourceName p) (sourceLine p) (sourceColumn p))))))
	where
		chainCalls clexpr =
			skipMany blank >>
			(char '(' >>
			sepBy parseRExpr (skipMany blank >> char ',') >>= (\args ->
				skipMany blank >>
				char ')' >>
				getPosition >>= (\p ->
					chainCalls (Call clexpr args (REPos (sourceName p) (sourceLine p) (sourceColumn p)))))) <||> 
			return clexpr
	
parseLocalBinds :: GenParser Char () RExpr
parseLocalBinds =
	skipMany blank >> 
	char '{' >>
	skipMany blank >>
	many1 (try parseFunction <|> parseValbind) >>= (\nfuncs ->
		skipMany blank >>
		char '}' >>
		skipMany blank >>
		parseRExpr >>= (\bexpr ->
			return (LocalBinds nfuncs bexpr))) 

parseCase :: GenParser Char () RExpr
parseCase =
		skipMany blank >> 
		getPosition >>= (\p ->
			(try parseCall <|> try parseLocalBinds <|> try parseLiteral <|> parseVar) >>= (\comparedExpr ->
				skipMany blank >>
				char '{' >>
				many1 branch >>= (\branches ->
					skipMany blank >>
					char '}' >>
					chainCase (Case comparedExpr branches (REPos (sourceName p) (sourceLine p) (sourceColumn p))))))
	where
		branch = 
			skipMany blank >>
			parseRExpr >>= (\x0 ->
				skipMany blank >>
				char ':' >>
				skipMany blank >>
				parseRExpr >>= (\x1 ->
					skipMany blank >>
					char ';' >>
					skipMany blank >>
					return (x0,x1)))
		chainCase caseexpr =
			skipMany blank >>
			getPosition >>= (\p ->
				(char '{' >>
				many1 branch >>= (\branches ->
					skipMany blank >>
					char '}' >>
					chainCase (Case caseexpr branches (REPos (sourceName p) (sourceLine p) (sourceColumn p))))) <|> 
				return caseexpr)
	
parseRExpr :: GenParser Char () RExpr 
parseRExpr = 
	try parseRFloat 		<|>
	try parseRInt  			<|>
	try parseRBool  		<|>
	try parseRChar  		<|>
	try parseRString 		<|>
	try parseRUnit  		<|>
	try parseRTuple 		<|>
	try parseRList 			<|>
	try parseRMap 			<|> 
	try parseCase 			<|>
	try parseCall 			<|>
	try parseRLambda		<|>  
	try parseLocalBinds <|> 
	parseVar

parseFunction :: GenParser Char () (String,RVal) 
parseFunction =
	skipMany blank >>
	many1 (noneOf "(){}[]=;:, ") >>= (\fname ->
		skipMany blank >>
		char '(' >> sepBy (skipMany blank >> many1 (noneOf "(){}[]=;:, ")) (skipMany blank >> char ',') >>= (\args ->
			char ')' >>
			skipMany blank >>
			char '=' >>
			parseRExpr >>= (\expr ->
				skipMany blank >> char ';' >> skipMany blank >> return (fname,RLambda args expr ML.empty))))
				
parseValbind :: GenParser Char () (String,RVal) 
parseValbind =
	skipMany blank >>
	many1 (noneOf "(){}[]=;:, ") >>= (\fname ->
		skipMany blank >>
		char '=' >>
		parseRExpr >>= (\expr ->
			skipMany blank >> 
			char ';' >> 
			skipMany blank >> 
			return (fname,RLambda [":RVALBIND"] expr ML.empty)))
				
parseRose :: String -> String -> Either String [(String,RVal)]
parseRose fname fcontent =
	case parse (manyTill parseFunction (try (skipMany blank >> eof))) fname fcontent of
		Right x -> Right x
		Left x -> Left (show x)

parseRoseExpr :: String -> Either String RExpr
parseRoseExpr content =
	case parse parseRExpr "(console)" content of
		Right x -> Right x
		Left x -> Left (show x)
