module Rose.Types (
	RVal(..),
	RExpr(..),
	REPos(..),
	valType,
	deepValType,
	listType,
	deepListType,
	fullListTypes,
	unpackRBool,
	unpackRBools,
	unpackRInt,
	unpackRInts,
	unpackRFloat,
	unpackRFloats,
	unpackRChar,
	unpackRChars,
	defaultREPos,
	unpackString) where
	
import Data.List as L
import Data.Map.Lazy as ML

import Rose.Misc

data RVal = 
	RBool Bool 																																							| 
	RInt Int 																																								| 
	RFloat Double 																																					| 
	RChar Char 																																							|
	RUnit																																										| 
	RTuple [RVal] 																																					| 
	RList [RVal] 																																						|
	RMap [(RVal,RVal)]																																			|
	RLambda {-Args names-}[String] {-Body-}RExpr {-local scope-}(ML.Map String RVal)				deriving(Show,Read,Eq)		

data REPos =
	REPos {-file-}String {-line-}Int {-column-}Int																					deriving(Show,Read,Eq)

defaultREPos = REPos ":rose-internal" (-1) (-1)
	
data RExpr =
	Literal {-Immediate value-}RVal 																												|
	Var {-Name in scope-}String {-position in code-}REPos																		|
	Call {-Expr which is lambda-}RExpr {-Args exprs-}[RExpr] {-position in code-}REPos			|
	LocalBinds {-Binds-}[(String,RVal)] {-Body-}RExpr																				|
	Case {-Expr for test-}RExpr {-Branches-}[(RExpr,RExpr)] {-position in code-}REPos				deriving(Show,Read,Eq)

valType :: RVal -> String
valType v =
	case v of
		RBool _ -> "RBool"
		RInt _ -> "RInt"
		RFloat _ -> "RFloat"
		RChar _ -> "RChar"
		RUnit -> "RUnit"
		RTuple _ -> "RTuple"
		RList _ -> "RList"
		RMap _ -> "RMap"
		RLambda _ _ _ -> "RLambda"

deepValType :: RVal -> String
deepValType val =
	case valType val of
		"RTuple" ->
			let
				RTuple innerList = val
			in
				"(" ++ L.tail (L.foldl (\a x -> a ++ "," ++ deepValType x) "" innerList) ++ ")"
		"RList" ->
			let
				RList innerList = val
				ilistType = listType innerList
				allTypeStructs = L.map (\x -> deepValType x) innerList
			in
				if ilistType == "error"
				then "[error]"
				else if ilistType == ""
				then "[]"
				else
					if (L.foldl (\a x -> if a == x then a else "ERROR") (L.head allTypeStructs) (L.tail allTypeStructs)) == "ERROR"
					then "[error]"
					else 
						"[" ++ L.head allTypeStructs ++ "]"
		"RMap" ->
			let
				RMap innerLists = val
				flist = RList (fst (L.unzip innerLists))
				slist = RList (snd (L.unzip innerLists))
			in
				"[" ++ L.init (L.tail (deepValType flist)) ++ ":" ++ L.init (L.tail (deepValType slist)) ++ "]"
		rest -> rest

listType :: [RVal] -> String
listType [] = ""
listType vals =
	L.foldl (\prev val ->
		if valType val == prev
		then prev
		else "error") (valType (L.head vals)) (L.tail vals)

deepListType :: [RVal] -> Bool
deepListType list =
	let
		dtypes = L.map (\x -> deepValType x) list 
	in
		if search "error" (L.head dtypes)
		then
			False
		else
			L.foldl (\a x -> if search "error" x then "ERROR" else if a == x then a else "ERROR") (L.head dtypes) (L.tail dtypes) /= "ERROR"

fullListTypes :: [RVal] -> String
fullListTypes list =
	"[" ++ L.concat (L.intersperse "," (L.map deepValType list)) ++ "]"
	 
unpackRBool :: RVal -> Bool
unpackRBool (RBool v) = v

unpackRBools :: [RVal] -> [Bool]
unpackRBools xs =
	L.map (\(RBool v) -> v) xs

unpackRInt :: RVal -> Int
unpackRInt (RInt v) = v

unpackRInts :: [RVal] -> [Int]
unpackRInts xs =
	L.map (\(RInt v) -> v) xs

unpackRFloat :: RVal -> Double
unpackRFloat (RFloat v) = v

unpackRFloats :: [RVal] -> [Double]
unpackRFloats xs =
	L.map (\(RFloat v) -> v) xs

unpackRChar :: RVal -> Char
unpackRChar (RChar v) = v

unpackRChars :: [RVal] -> String
unpackRChars xs =
	L.map (\(RChar v) -> v) xs
	
unpackString :: RVal -> Maybe String
unpackString val =
	if valType val == "RList"
	then
		let
			RList ilist = val
		in
			if listType ilist == "RChar"
			then Just $ L.map (\(RChar c) -> c) ilist
			else Nothing
	else Nothing
